package cn.com.qmth.stmms.ms.marking.service;

import cn.com.qmth.stmms.ms.commons.utils.RandomUtil;
import cn.com.qmth.stmms.ms.core.domain.MarkStage;
import cn.com.qmth.stmms.ms.core.domain.MarkSubject;
import cn.com.qmth.stmms.ms.core.domain.Paper;
import cn.com.qmth.stmms.ms.core.domain.task.MarkTask;
import cn.com.qmth.stmms.ms.core.domain.user.MarkUser;
import cn.com.qmth.stmms.ms.core.domain.user.MarkerGroup;
import cn.com.qmth.stmms.ms.core.repository.MarkSubjectRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkTaskRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkerGroupRepo;
import cn.com.qmth.stmms.ms.core.repository.PaperRepo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.*;
import java.util.stream.Collectors;

/**
 * 评卷任务分配
 * Created by zhengmin on 2016/9/23.
 */
@Service
public class AssignTaskService {

    @Autowired
    private MarkTaskRepo markTaskRepo;

    @Autowired
    private PaperRepo paperRepo;

    @Autowired
    private MarkSubjectRepo markSubjectRepo;

    @Autowired
    private MarkerGroupRepo markerGroupRepo;

    @Value("${sys.config.random-bundle-size}")
    private int randomBundleSize;


    /**
     * 单评任务模式，每个科目评卷员都需要对该科目每份试卷进行评卷
     * 比如，色彩有1000份试卷，那么每个评卷员都需要评1000份
     *
     * @param papers
     * @param markers
     */
    @Transactional
    public void assignForAll(List<Paper> papers, Iterable<MarkUser> markers, MarkSubject markSubject) throws Exception {
        List<MarkTask> markTasks = new ArrayList<>();
        int paperCount = papers.size();
        int randomBundleCount = paperCount / randomBundleSize;
        if (paperCount % randomBundleSize != 0) {
            randomBundleCount += 1;
        }

        for (int i = 0; i < randomBundleCount; i++) {
            int fromIndex = i * randomBundleSize;
            int endIndex = i * randomBundleSize + randomBundleSize;
            if (endIndex > paperCount) {
                endIndex = paperCount;
            }
            List<Paper> bundleList = papers.subList(fromIndex, endIndex);

            for (MarkUser marker : markers) {
                List<Paper> markerPaperList = Arrays.asList(new Paper[bundleList.size()]);

                Collections.copy(markerPaperList, bundleList);
                Collections.shuffle(markerPaperList);//乱序

                for (int j = 0; j < markerPaperList.size(); j++) {
                    Paper paper = markerPaperList.get(j);
                    Long random = getRandom(marker.getId(), paper.getId(), marker.getWorkId(), paper.getExamNumber());
                    MarkTask markTask = new MarkTask(marker, paper, markSubject.getStage(), random, markSubject.getTest());
                    markTask.setRandomSeq(i * randomBundleSize + j + 1);
                    markTasks.add(markTask);
                }
            }

        }

        markTaskRepo.save(markTasks);

//        Iterator<Paper> iterator = papers.iterator();
//        while (iterator.hasNext()) {
//            Paper paper = iterator.next();
//            for (MarkUser marker : markers) {
//                MarkTask markTask = new MarkTask(marker, paper, markSubject.getStage());
//                markTaskRepo.save(markTask);
//                //taskExecutor.execute(papers,marker,stage);
//            }
//        }
    }

    /**
     * 分组任务模式，将该科目的评卷员分成N组，将该科目试卷平均分给这些组，每组内的评卷员每人将这些试卷打分
     * 比如：6个评卷员分2组，有1000份试卷，每组评500份，每个评卷员评500份
     *
     * @param papers
     * @param markerGroups
     */
    @Transactional
    public void assignForGrouping(List<Paper> papers, List<MarkerGroup> markerGroups, MarkSubject markSubject) throws Exception {
        markerGroups = markerGroups.stream().filter(m -> m.getMarkers().size() > 0).collect(Collectors.toList());
        int groupSize = markerGroups.size();
        int idx = 0;
        long currentTime = System.currentTimeMillis();
        Iterator<Paper> iterator = papers.iterator();
        while (iterator.hasNext()) {
            if (idx >= groupSize) {
                idx = 0;
            }
            Paper paper = iterator.next();
            for (MarkUser marker : markerGroups.get(idx).getMarkers()) {
                Long random = getRandom(marker.getId(), paper.getId(), marker.getWorkId(), paper.getExamNumber());
                MarkTask markTask = new MarkTask(marker, paper, markSubject.getStage(), random, markSubject.getTest());
                markTaskRepo.save(markTask);
            }
            if (markSubject.getTest() != 1) {
                if (MarkStage.LEVEL.equals(markSubject.getStage())) {
                    //将paper置为激活
                    paper.setActive(true);
                    paper.setBatchNo(currentTime);
                    paperRepo.save(paper);

                    //更新subject表formal为正试评卷
                    markSubject.setFormal(true);
                    markSubjectRepo.saveAndFlush(markSubject);
                }
            }
            idx++;
        }
        //只有全部任务投放完成，才能删除分组
        int count = paperRepo.countByWorkIdAndSubjectAndIsMissingFalseAndActiveFalseAndBatchNoIsNullAndTest(markSubject.getWorkId(), markSubject.getSubject(), 0);
        //进入分档阶段后，清空当前分组，以便打分阶段前重新分组
        if (MarkStage.LEVEL.equals(markSubject.getStage()) && count == 0) {
            for (MarkerGroup markerGroup : markerGroups) {
                markerGroup.setMarkers(null);
            }
            markerGroupRepo.delete(markerGroups);
        }
    }

    /**
     * 获取随机号
     *
     * @param markerId
     * @param paperId
     * @param workId
     * @param examNumber
     * @return
     * @throws Exception
     */
    public Long getRandom(Long markerId, Long paperId, Long workId, String examNumber) throws Exception {
        int count = 0, result = 0;
        Long random = 0L;
        while (true) {
            random = RandomUtil.randomMap.get(workId).get(new Random().nextInt(RandomUtil.randomMap.get(workId).size()));
//            random = RandomUtil.randomList.get(new Random().nextInt(RandomUtil.randomList.size()));
            result = markTaskRepo.countByMarkerIdAndPaperIdAndWorkIdAndRandomSeqNew(markerId, paperId, workId, random);
            if (result == 0 && random != Long.parseLong(examNumber.substring(3, examNumber.length()))) {
                break;
            } else {
                count++;
            }
            if (count > 100) {
//                throw new Exception("重复几率较高，建议重新生成随机号");
                RandomUtil.getRandom(workId, true);
            }
        }
        return random;
    }
}
