package cn.com.qmth.stmms.ms.marking.service;

import cn.com.qmth.stmms.ms.core.domain.*;
import cn.com.qmth.stmms.ms.core.domain.enums.TrialEnum;
import cn.com.qmth.stmms.ms.core.domain.user.MarkRight;
import cn.com.qmth.stmms.ms.core.domain.user.MarkUser;
import cn.com.qmth.stmms.ms.core.domain.user.MarkerGroup;
import cn.com.qmth.stmms.ms.core.domain.user.Role;
import cn.com.qmth.stmms.ms.core.repository.*;
import cn.com.qmth.stmms.ms.core.vo.Subject;
import cn.com.qmth.stmms.ms.marking.dto.LevleProgressDTO;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Sort;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 评卷阶段控制服务
 * Created by zhengmin on 2016/9/24.
 */
@Service
public class StageControlService {

    @Autowired
    private MarkSubjectRepo markSubjectRepo;

    @Autowired
    private MarkTaskRepo markTaskRepo;

    @Autowired
    private LevelRepo levelRepo;

    @Autowired
    private PaperRepo paperRepo;

    @Autowired
    private MarkUserRepo markUserRepo;

    @Autowired
    private AssignTaskService assignTaskService;

    @Autowired
    private ExamQuestionRepo examQuestionRepo;

    @Autowired
    private MarkerGroupRepo markerGroupRepo;

    public static volatile Map<String, Boolean> statusMap = new Hashtable<>();

    /**
     * 进入下一阶段
     *
     * @param markSubject
     */
    @Transactional
    public void goNext(MarkSubject markSubject, int taskCount) throws Exception {
        if (statusMap.get(markSubject.getId()) != null && statusMap.get(markSubject.getId())) {
            return;
        }
        statusMap.put(markSubject.getId(), true);
        MarkStage markStage = markSubject.getStage();
        //是否还有未投放的试卷
        int waitCount = paperRepo.countByWorkIdAndSubjectAndIsMissingFalseAndActiveFalseAndBatchNoIsNullAndTest(markSubject.getWorkId(), markSubject.getSubject(), 0);
        if (waitCount > 0 && markStage.equals(MarkStage.LEVEL)) {
            markStage = MarkStage.INIT;
        }
        switch (markStage) {
            case INIT:
                markSubject.setStage(MarkStage.LEVEL);
                enterLevelStage(markSubject, taskCount);
                break;
            case LEVEL:
                markSubject.setStage(MarkStage.SCORE);
                enterScoreStage(markSubject);
                break;
            case SCORE:
                //TODO 进入抽查阶段
                markSubject.setStage(MarkStage.INSPECT);
                enterInspectStage(markSubject);
                break;
            case INSPECT:
                //TODO 结束评卷
                markSubject.setStage(MarkStage.END);
                break;
        }
        markSubjectRepo.save(markSubject);
        statusMap.put(markSubject.getId(), false);
    }

    /**
     * 进入分档阶段
     * 1.检查是否有设定档位，如果有，将所有level设置连续的levelValue值，用于仲裁误差判定
     * 2.检查有无可分档的评卷员，没有则终止
     * 3.给所有能够分档的评卷员分配任务
     *
     * @param markSubject
     */
    private void enterLevelStage(MarkSubject markSubject, int taskCount) throws Exception {
        Long workId = markSubject.getWorkId();
        Subject subject = markSubject.getSubject();
        List<Level> levels = levelRepo.findByWorkId(workId);
        if (levels.size() == 0) {
            statusMap.put(markSubject.getId(), false);
            throw new RuntimeException("没有设定档位");
        }

        List<MarkUser> markers = markUserRepo.findByWorkIdAndSubjectAndRole(workId, subject, Role.MARKER);
        List<MarkUser> levelMarkers = markers.stream().filter(m -> m.getMarkRight() == MarkRight.ALLOW_ALL || m.getMarkRight() == MarkRight.ALLOW_LEVELING).collect(Collectors.toList());
        if (levelMarkers.size() == 0) {
            statusMap.put(markSubject.getId(), false);
            throw new RuntimeException("没有设定评卷员");
        }
        for (int i = 0; i < levels.size(); i++) {
            levels.get(i).setLevelValue(i);
        }
        levelRepo.save(levels);
        //assignTaskService.assignForAll(papers, levelMarkers, markSubject);
        List<MarkerGroup> markerGroups = markerGroupRepo.findByWorkIdAndSubject(workId, subject);
        if (markerGroups.size() == 0) {
            statusMap.put(markSubject.getId(), false);
            throw new RuntimeException("评卷员没有进行分组");
        }

        if (markSubject.getTest() != 1) {
            //是否有未结束的分档任务
            int count = paperRepo.countByWorkIdAndSubjectAndLevelIsNullAndIsMissingFalseAndActiveTrueAndBatchNoNotNullAndTest(workId, subject, 0);
            if (count > 0) {
                statusMap.put(markSubject.getId(), false);
                throw new RuntimeException("没有分档完成");
            }

            //任务数量大于0，小于等于待评数量
            if (taskCount == 0) {
                statusMap.put(markSubject.getId(), false);
                throw new RuntimeException("请输入正确的分配任务数量");
            }
            int waitCount = paperRepo.countByWorkIdAndSubjectAndIsMissingFalseAndActiveFalseAndBatchNoIsNullAndTest(workId, subject, 0);
            if (taskCount - waitCount > 0) {
                statusMap.put(markSubject.getId(), false);
                throw new RuntimeException("分配任务数量不能大于待评数量");
            }
        }

        List<Paper> papers = null;
        if (markSubject.getTest() == 1) {
            papers = paperRepo.findByWorkIdAndSubjectAndIsMissingAndTest(workId, subject, false, markSubject.getTest());
        } else {
            List<Paper> papersList = paperRepo.findByWorkIdAndSubjectAndIsMissingFalseAndActiveFalseAndTest(workId, subject, TrialEnum.DEFAULT.getId());
            papers = papersList.subList(0, taskCount);
        }
        assignTaskService.assignForGrouping(papers, markerGroups, markSubject);
    }

    /**
     * 进入打分阶段
     * 1.检查是否有能够打分的评卷员
     * 2.检查是否所有试卷都分档完成（有仲裁卷或打回重评卷）
     *
     * @param markSubject
     */
    private void enterScoreStage(MarkSubject markSubject) throws Exception {
        Long workId = markSubject.getWorkId();
        Subject subject = markSubject.getSubject();
        List<MarkUser> markers = markUserRepo.findByWorkIdAndSubjectAndRole(workId, subject, Role.MARKER);
        List<MarkUser> levelMarkers = markers.stream().filter(m -> m.getMarkRight() == MarkRight.ALLOW_ALL || m.getMarkRight() == MarkRight.ALLOW_SCORING).collect(Collectors.toList());
        if (levelMarkers.size() == 0) {
            statusMap.put(markSubject.getId(), false);
            throw new RuntimeException("没有设定评卷员");
        }
        int count = paperRepo.countByWorkIdAndSubjectAndLevelIsNullAndIsMissingFalseAndActiveTrueAndBatchNoNotNullAndTest(workId, subject, 0);
        if (count > 0) {
            statusMap.put(markSubject.getId(), false);
            throw new RuntimeException("没有分档完成");
        }
        this.isPtFull(workId, markSubject);
        Sort sort = new Sort("questionId");
        List<Paper> papers = paperRepo.findByWorkIdAndSubjectAndIsMissingFalseAndActiveTrue(workId, subject, sort);
        List<MarkerGroup> markerGroups = markerGroupRepo.findByWorkIdAndSubject(workId, subject);
        if (markerGroups.size() == 0) {
            statusMap.put(markSubject.getId(), false);
            throw new RuntimeException("评卷员没有进行分组");
        }
        assignTaskService.assignForGrouping(papers, markerGroups, markSubject);
    }

    /**
     * @param workId
     * @param markSubject
     */
    private void isPtFull(Long workId, MarkSubject markSubject) {
        //控制分档比例
        List<Level> levels = levelRepo.findByWorkId(workId);
        for (Level targetLevel : levels) {
            int pt = targetLevel.getPt();
            long count = paperRepo.countByWorkIdAndSubjectAndLevel(workId, markSubject.getSubject(), targetLevel.getCode());
            long total = paperRepo.countByWorkIdAndSubject(workId, markSubject.getSubject());

            if ((count * 1.0D) / total * 1.0D > pt / 100.D) {
                statusMap.put(markSubject.getId(), false);
                throw new RuntimeException(targetLevel.getCode() + "档已满，占比总阀值已达" + pt + "%。");
            }
            //判断考点阀值是否已满
            int kdpt = targetLevel.getKdpt();
            List<ExamQuestion> examQuestions = examQuestionRepo.findByWorkIdAndSubject(workId, markSubject.getSubject());
            for (ExamQuestion examQuestion : examQuestions) {
                long kdcount = paperRepo.countByWorkIdAndSubjectAndQuestionIdAndLevel(workId, markSubject.getSubject(), examQuestion.getId(), targetLevel.getCode());
                long kdtotal = paperRepo.countByWorkIdAndSubjectAndQuestionId(workId, markSubject.getSubject(), examQuestion.getId());
                if ((kdcount * 1.0D) / kdtotal * 1.0D > kdpt / 100.D) {
                    statusMap.put(markSubject.getId(), false);
                    throw new RuntimeException(targetLevel.getCode() + "档已满，考点阀值已达" + kdpt + "%。");
                }
            }
        }
    }

    /**
     * 进入到抽查阶段
     *
     * @param markSubject
     */
    private void enterInspectStage(MarkSubject markSubject) {
        // TODO: 2017/10/11
        Long workId = markSubject.getWorkId();
        Subject subject = markSubject.getSubject();
//        List<MarkUser> checkers = markUserRepo.findByWorkIdAndSubjectAndRole(workId, subject, Role.INSPECTOR);
//        if (checkers.size() == 0) {
//            statusMap.put(markSubject.getId(), false);
//            throw new RuntimeException("没有设定复评专家");
//        }

    }

    /**
     * 重置所有评卷数据，回到INIT状态
     *
     * @param markSubject
     * @param stage       重置的阶段
     * @param force       是否强制重置，如果为false，当有评卷记录时，不可重置，如果为true，不论有无评卷记录，删除所有数据
     */
    @Transactional
    public void reset(MarkSubject markSubject, MarkStage stage, boolean force) {
        Subject subject = markSubject.getSubject();
        Long workId = markSubject.getWorkId();
        markTaskRepo.deleteByWorkIdAndSubjectAndStage(workId, subject, markSubject.getStage());
        markSubject.setStage(stage);
        markSubjectRepo.save(markSubject);
    }

    /**
     * 分档进度查询
     *
     * @param markSubject
     */
    public LevleProgressDTO progress(MarkSubject markSubject) {
        Long workId = markSubject.getWorkId();
        Subject subject = markSubject.getSubject();
        //已评任务数
        int successCount = paperRepo.countByWorkIdAndSubjectAndLevelNotNullAndIsMissingFalseAndActiveTrueAndTest(workId, subject, 0);
        //所有任务数(查询所有有效试卷)
        int totalCount = paperRepo.countByWorkIdAndSubjectAndIsMissingFalseAndTest(workId, subject, 0);

        BigDecimal waitTotal = BigDecimal.valueOf(totalCount).subtract(BigDecimal.valueOf(successCount));
        //进度
        BigDecimal progress = BigDecimal.ZERO.compareTo(BigDecimal.valueOf(totalCount)) == 0 ? BigDecimal.ZERO :BigDecimal.valueOf(successCount).multiply(BigDecimal.valueOf(100)).divide(BigDecimal.valueOf(totalCount), 2, BigDecimal.ROUND_HALF_UP);

        LevleProgressDTO progressDTO = new LevleProgressDTO();
        progressDTO.setSuccessCount(successCount);
        progressDTO.setTotalCount(totalCount);
        progressDTO.setWaitCount(waitTotal.intValue());
        progressDTO.setProgress(progress.doubleValue());

        return progressDTO;
    }

}
